unit UCore;

{$mode objfpc}{$H+}

interface

uses
  Classes, SysUtils, Contnrs, Graphics;

type
  TEngine = class;

  TPlayerKeys = record
    Left: Word;
    Right: Word;
    Up: Word;
    Down: Word;
    Shoot: Word;
  end;

  { TPlayer }

  TPlayer = class
    Engine: TEngine;
    Color: TColor;
    Position: TPoint;
    ScreenFrame: TRect;
    Name: string;
    Keys: TPlayerKeys;
    procedure Control;
    procedure Paint;
  end;

  { TWorld }

  TWorld = class
  private
    function GetSize: TPoint;
    procedure SetSize(const AValue: TPoint);
  public
    Engine: TEngine;
    Surface: TBitmap;
    procedure Generate;
    constructor Create;
    destructor Destroy; override;
    property Size: TPoint read GetSize write SetSize;
  end;

  { TEngine }

  TEngine = class
  private
    FBitmap: TBitmap;
    function GetPlayerCount: Integer;
    procedure SetBitmap(const AValue: TBitmap);
    procedure SetPlayerCount(const AValue: Integer);
  public
    KeyState: array[0..High(Word)] of Boolean;
    World: TWorld;
    Players: TObjectList; // <TPlayer>
    constructor Create;
    destructor Destroy; override;
    procedure ResizePlayerFrames;
    procedure Paint;
    property PlayerCount: Integer read GetPlayerCount write SetPlayerCount;
    property Bitmap: TBitmap read FBitmap write SetBitmap;
  end;

var
  Engine: TEngine;

implementation

{ TWorld }

function TWorld.GetSize: TPoint;
begin
  Result := Point(Surface.Width, Surface.Height);
end;

procedure TWorld.SetSize(const AValue: TPoint);
begin
  Surface.SetSize(AValue.X, AValue.Y);
  Generate;
end;

procedure TWorld.Generate;
var
  X, Y: Integer;
begin
  for Y := 0 to Surface.Height - 1 do
    for X := 0 to Surface.Width - 1 do begin
      if Random < 0.5 then
        Surface.Canvas.Pixels[X, Y] := TColor($0756b0) else
        Surface.Canvas.Pixels[X, Y] := TColor($2170c3);
    end;
end;

constructor TWorld.Create;
begin
  Surface := TBitmap.Create;
  Surface.PixelFormat := pf4bit;
  Size := Point(500, 100);
end;

destructor TWorld.Destroy;
begin
  Surface.Free;
  inherited Destroy;
end;

{ TPlayer }

procedure TPlayer.Control;
begin
  if Engine.KeyState[Ord(Keys.Up)] then Position.Y := Position.Y + 1;
  if Engine.KeyState[Ord(Keys.Down)] then Position.Y := Position.Y - 1;
  if Engine.KeyState[Ord(Keys.Right)] then Position.X := Position.X + 1;
  if Engine.KeyState[Ord(Keys.Left)] then Position.X := Position.X - 1;
end;

procedure TPlayer.Paint;
begin
  with Engine.Bitmap.Canvas do begin
    Rectangle(ScreenFrame);
    CopyRect(ScreenFrame, Engine.World.Surface.Canvas, Rect(
      Position.X - (ScreenFrame.Right - ScreenFrame.Left) div 2,
      Position.Y - (ScreenFrame.Bottom - ScreenFrame.Top) div 2,
      Position.X + (ScreenFrame.Right - ScreenFrame.Left) div 2,
      Position.Y + (ScreenFrame.Bottom - ScreenFrame.Top) div 2));
    TextOut(ScreenFrame.Left, ScreenFrame.Top, Name);
  end;
end;

{ TEngine }

function TEngine.GetPlayerCount: Integer;
begin
  Result := Players.Count;
end;

procedure TEngine.SetBitmap(const AValue: TBitmap);
begin
  FBitmap := AValue;
  ResizePlayerFrames;
end;

procedure TEngine.SetPlayerCount(const AValue: Integer);
var
  NewPlayer: TPlayer;
  I: Integer;
begin
  if Players.Count < AValue then begin
    for I := Players.Count to AValue - 1 do begin
      NewPlayer := TPlayer.Create;
      NewPlayer.Engine := Self;
      NewPlayer.Name := 'Player ' + IntToStr(I);
      Players.Add(NewPlayer);
    end;
  end else
  if Players.Count > AValue then begin
    for I := Players.Count - 1 downto AValue do
      Players.Delete(I);
  end;
  ResizePlayerFrames;
end;

procedure TEngine.ResizePlayerFrames;
var
  HorizFrameCount: Integer;
  VertFrameCount: Integer;
  I: Integer;
begin
  // TODO: Determine frames from player count
  if Assigned(Bitmap) then begin
    HorizFrameCount := 2;
    VertFrameCount := 1;
    for I := 0 to Players.Count - 1 do begin
      TPlayer(Players[I]).ScreenFrame := Rect(
        (I mod HorizFrameCount) * (Bitmap.Width div HorizFrameCount),
        (I div HorizFrameCount) * (Bitmap.Height div VertFrameCount),
        ((I mod HorizFrameCount) + 1) * (Bitmap.Width div HorizFrameCount),
        ((I div HorizFrameCount) + 1) * (Bitmap.Height div VertFrameCount));
    end;
  end;
end;

constructor TEngine.Create;
begin
  Players := TObjectList.Create;
  PlayerCount := 2;
  with TPlayer(Players[0]) do begin
    Position := Point(100, 100);
    Color := TColor($00f805);
    Keys.Left := 65;
    Keys.Up := 87;
    Keys.Right := 68;
    Keys.Down := 83;
    Keys.Shoot := 69;
  end;
  with TPlayer(Players[1]) do begin
    Position := Point(300, 200);
    Color := TColor($00f805);
    Keys.Left := 37;
    Keys.Up := 38;
    Keys.Right := 39;
    Keys.Down := 40;
    Keys.Shoot := 17;
  end;
  World := TWorld.Create;
  World.Engine := Self;
end;

destructor TEngine.Destroy;
begin
  Players.Free;
  World.Free;
  inherited Destroy;
end;

procedure TEngine.Paint;
var
  I: Integer;
begin
  if Assigned(Bitmap) then begin
    Bitmap.Canvas.FillRect(0, 0, Bitmap.Width, Bitmap.Height);
    for I := 0 to Players.Count - 1 do begin
      TPlayer(Players[I]).Control;
      TPlayer(Players[I]).Paint;
    end;
  end;
end;

initialization

Engine := TEngine.Create;

finalization

Engine.Free;

end.

